from odoo import _
from odoo.exceptions import MissingError

from odoo.addons.base_rest import restapi
from odoo.addons.base_rest_datamodel.restapi import Datamodel
from odoo.addons.component.core import Component

from ..pms_api_rest_utils import pms_api_check_access


class PmsBoardServiceService(Component):
    _inherit = "base.rest.service"
    _name = "pms.board.service.service"
    _usage = "board-services"
    _collection = "pms.services"

    @restapi.method(
        [
            (
                [
                    "/",
                ],
                "GET",
            )
        ],
        input_param=Datamodel("pms.board.service.search.param"),
        output_param=Datamodel("pms.board.service.info", is_list=True),
        auth="jwt_api_pms",
    )
    def get_board_services(self, board_services_search_param):
        external_app = self.env.user.pms_api_client
        domain = []
        if board_services_search_param.name:
            domain.append(("name", "like", board_services_search_param.name))
        if board_services_search_param.roomTypeId:
            domain.append(
                ("pms_room_type_id", "=", board_services_search_param.roomTypeId)
            )
        if board_services_search_param.pmsPropertyId:
            domain.extend(
                [
                    ("pms_property_id", "=", board_services_search_param.pmsPropertyId),
                ]
            )

        result_board_services = []
        PmsBoardServiceInfo = self.env.datamodels["pms.board.service.info"]
        board_services = self.env["pms.board.service.room.type"].sudo().search(domain)
        pms_api_check_access(user=self.env.user, records=board_services)
        for board_service in board_services:
            result_board_services.append(
                PmsBoardServiceInfo(
                    id=board_service.id,
                    name=board_service.pms_board_service_id.display_name,
                    roomTypeId=board_service.pms_room_type_id.id,
                    amount=round(board_service.amount, 2),
                    boardServiceId=board_service.pms_board_service_id,
                    boardServiceLineIds=board_service.board_service_line_ids.ids,
                )
            )
        if external_app:
            room_type_ids = board_services_search_param.roomTypeId or self.env[
                "pms.room"
            ].sudo().search(
                [("pms_property_id", "=", board_services_search_param.pmsPropertyId)]
            ).mapped(
                "room_type_id.id"
            )
            pms_api_check_access(user=self.env.user, records=room_type_ids)
            for room_type_id in room_type_ids:
                result_board_services.append(
                    PmsBoardServiceInfo(
                        id=0,
                        name="Solo Alojamiento",
                        roomTypeId=room_type_id,
                        amount=0,
                        boardServiceId=0,
                        boardServiceLineIds=[],
                    )
                )
        return result_board_services

    @restapi.method(
        [
            (
                [
                    "/<int:board_service_id>",
                ],
                "GET",
            )
        ],
        output_param=Datamodel("pms.board.service.info", is_list=False),
        auth="jwt_api_pms",
    )
    def get_board_service(self, board_service_id):
        board_service = (
            self.env["pms.board.service.room.type"]
            .sudo()
            .search([("id", "=", board_service_id)])
        )
        if board_service:
            pms_api_check_access(user=self.env.user, records=board_service)
            PmsBoardServiceInfo = self.env.datamodels["pms.board.service.info"]
            return PmsBoardServiceInfo(
                id=board_service.id,
                name=board_service.pms_board_service_id.display_name,
                roomTypeId=board_service.pms_room_type_id.id,
                amount=round(board_service.amount),
                boardServiceLineIds=board_service.board_service_line_ids.ids,
            )
        else:
            raise MissingError(_("Board Service not found"))

    @restapi.method(
        [
            (
                [
                    "/<int:board_service_id>/lines",
                ],
                "GET",
            )
        ],
        input_param=Datamodel("pms.search.param"),
        output_param=Datamodel("pms.board.service.line.info", is_list=True),
        auth="jwt_api_pms",
    )
    def get_board_service_lines(self, board_service_id, pms_search_param):
        domain = list()
        domain.append(("pms_board_service_room_type_id", "=", board_service_id))
        if pms_search_param.pmsPropertyId:
            domain.extend(
                [
                    ("pms_property_id", "=", pms_search_param.pmsPropertyId),
                ]
            )
        result_board_service_lines = []
        PmsBoardServiceLineInfo = self.env.datamodels["pms.board.service.line.info"]
        lines = self.env["pms.board.service.room.type.line"].sudo().search(domain)
        pms_api_check_access(user=self.env.user, records=lines)
        for line in lines:
            result_board_service_lines.append(
                PmsBoardServiceLineInfo(
                    id=line.id,
                    name=line.pms_board_service_room_type_id.display_name,
                    boardServiceId=line.pms_board_service_room_type_id.id,
                    productId=line.product_id.id,
                    amount=round(line.amount, 2),
                    isAdults=line.adults,
                    isChildren=line.children,
                )
            )
        return result_board_service_lines

    @restapi.method(
        [
            (
                [
                    "/restricted/<int:board_service_id>",
                ],
                "GET",
            )
        ],
        output_param=Datamodel("pms.board.service.info", is_list=False),
        auth="jwt_api_pms",
    )
    def get_restricted_board_service(self, board_service_id):
        board_service_record = (
            self.env["pms.board.service.room.type"].sudo().browse(board_service_id)
        )
        if board_service_record.exists():
            PmsBoardServiceInfo = self.env.datamodels["pms.board.service.info"]
            return PmsBoardServiceInfo(
                id=board_service_record.id,
                name=board_service_record.pms_board_service_id.display_name,
                roomTypeId=board_service_record.pms_room_type_id.id,
                amount=round(board_service_record.amount),
                productIds=board_service_record.board_service_line_ids.mapped(
                    "product_id.id"
                ),
            )
        else:
            raise MissingError(_("Board Service not found"))
